Unit AlgArit;
{ calculs arithmtiques pour CalcAlg }
{ K.B. 97 }

INTERFACE

Type Str12=String[12];

Var ArithmError:Word;  { codes d'erreurs }
Const
 NoErr=0;     { 0 : pas d'erreur }
 errCapa=1;   { 1 : dpassement de capacit }
 errDiv0=2;   { 2 : division par zro }
 errPosit=3;  { 3 : nombre positif attendu }
 errNum=4;    { 4 : nombre mal crit }

Function GetArithmErrorMsg:String;
{ renvoie le message d'erreur }

Function PGCD(n1,n2:LongInt):LongInt;
{ calcul du PGCD de 2 entiers }

Function PPCM(n1,n2:LongInt):LongInt;
{ calcul du PPCM de 2 entiers }

Function Bezout(n1,n2:LongInt;Var u1,u2:LongInt):LongInt;
{ calcule u1 et u2 tels que n1u1+n2u2=PGCD(n1,n2) }

Function PremDivSuiv(n,d:LongInt):LongInt;
{ calcul du premier diviseur premier de n suprieur  d }

Function PremDiv(n:LongInt):LongInt;
{ calcul du premier diviseur premier d'un nombre entier }

Procedure ToFrac(r:Real; var n,d:LongInt);
{ transforme un type Real en fraction }

Procedure SimpliFrac(var n,d:LongInt);
{ simplifie la fraction n/d }

Function FacteurCarre(n:LongInt):LongInt;
{ renvoie le plus grand facteur carr de n }

IMPLEMENTATION

Procedure ToFrac(r:Real; var n,d:LongInt);
Var w,wr:Real;
    numer1,denom1:longint;
    numer2,denom2:longint;
    p,q:longint;
    ok:Boolean;
Begin
 n:=0; d:=0;
 if abs(r)>=MaxLongInt then
  begin
   ArithmError:=errCapa;
   exit;
  end;
 ArithmError:=NoErr;
 wr:=abs(r);
 ok:=false;
 n:=trunc(wr);
 if wr=n then
  ok:=true
 else w:=1/(wr-n);
 numer1:=1;  denom1:=0;
 numer2:=n;  denom2:=1;
 p:=numer2;  q:=denom2;
 while not(ok) do
  begin
   if q<>0 then
    if abs(p/q-wr)<1E-8 then
     ok:=true
     else
      begin
       n:=trunc(w);
       p:=n*numer2+numer1;
       q:=n*denom2+denom1;
       numer1:=numer2; denom1:=denom2;
       numer2:=p; denom2:=q;
       if w=n then
        ok:=true
       else w:=1/(w-n);
      end;
  end;
 n:=p;
 if r<0 then n:=-n;
 d:=q;
End;

Procedure SimpliFrac(var n,d:LongInt);
Var e:LongInt;
Begin
 ArithmError:=NoErr;
 if d<0 then
  begin n:=-n; d:=-d end;
 e:=PGCD(n,d);
 if e<>0 then
  begin
   n:=n div e;
   d:=d div e;
  end
 else ArithmError:=errDiv0;
End;

Function GetArithmErrorMsg:String;
Begin
 case ArithmError of
  NoErr : GetArithmErrorMsg:='Calcul effectu sans erreur.';
  errCapa : GetArithmErrorMsg:='Dpassement de capacit.';
  errDiv0 : GetArithmErrorMsg:='Division par zro.';
  errPosit : GetArithmErrorMsg:='Nombre positif attendu.';
  errNum  : GetArithmErrorMsg:='Nombre mal crit.';
  else GetArithmErrorMsg:='???';
  end;
End;

Function PGCD(n1,n2:LongInt):LongInt;
{ Dtermination du Plus Grand Commun Diviseur de n1 et n2  par la
  mthode des divisions successives. }
Var a,b,r:longint;
Begin
 ArithmError:=NoErr;
 a:=n1;
 b:=n2;
 if a<b then
  begin r:=a; a:=b; b:=r end;
 if b=0 then
  begin PGCD:=a; exit end;
 repeat
  r:=a mod b;
  if r<>0 then
   begin a:=b; b:=r end;
 until r=0;
 PGCD:=b;
End;

Function PPCM(n1,n2:LongInt):LongInt;
Var w:LongInt;
Begin
 ArithmError:=NoErr;
 PPCM:=0;
 if (n1=0) or (n2=0) then exit;
 w:=n1 div PGCD(n1,n2);
 if MaxLongInt div n2 < w then
  begin
   ArithmError:=errCapa;
   exit;
  end
 else PPCM:=w*n2;
End;

Function Bezout(n1,n2:LongInt;Var u1,u2:LongInt):LongInt;
{ calcule u1 et u2 tels que n1u1+n2u2=PGCD(n1,n2) et renvoie le PGCD }
Var t,r1,r2,t1,t2:LongInt;
Begin
 if n1<n2 then
  begin
   t:=n1;
   n1:=n2; n2:=t;
   u1:=0;  u2:=1;
   r1:=1;  r2:=0;
  end
 else
  begin
   u1:=1;  u2:=0;
   r1:=0;  r2:=1;
  end;
 while n2<>0 do
  begin
   t:= n1 mod n2;
   t1:=u1-(n1 div n2)*r1;
   t2:=u2-(n1 div n2)*r2;
   n1:=n2; n2:=t;
   u1:=r1; u2:=r2;
   r1:=t1; r2:=t2;
  end;
 Bezout:=n1;
End;

Function PremDiv(n:LongInt):LongInt;
Begin
 if n>1 then
  PremDiv:=PremDivSuiv(n,2)
 else PremDiv:=1;
End;

Function PremDivSuiv(n,d:LongInt):LongInt;
{ recherche du premier diviseur premier suprieur ou gal  d }
Var dd,r:longint;
Begin
 if n<=d then
  begin
   PremDivSuiv:=n;
   exit;
  end;
 r:=n mod d;
 if r=0 then
  begin
   PremDivSuiv:=d;
   exit;
  end;
 if d=2 then
  begin
   d:=3;
   r:=n mod d;
   if r=0 then
    begin
     PremDivSuiv:=d;
     exit;
    end;
  end;
 if d=3 then d:=1; (* pour passer  5 en ajoutant 4 *)
 if d mod 6 = 1
  then dd:=4 else dd:=2;
 repeat
  inc(d,dd);
  r:=n mod d;
  dd:=6-dd;
 until (d*d>n) or (r=0);
 if r=0 then
  PremDivSuiv:=d
 else PremDivSuiv:=n;
End;

Function FacteurCarre(n:LongInt):LongInt;
Var p,d,r:LongInt;
Begin
 r:=1; d:=2;
 p:=PremDiv(n);
 repeat
  if p<>n then
   begin
    n:=n div p;
    if n mod p = 0 then
     begin
      r:=r*p;
      n:=n div p;
     end
    else
     begin
      if p=2 then p:=3
      else if p=3 then p:=5
      else if p mod 6=1 then inc(p,4)
      else inc(p,2);
     end;
    p:=PremDivSuiv(n,p);
   end;
 until p=n;
 FacteurCarre:=r;
End;

END.

{                        Fin du fichier ALGARIT.PAS                         }
