Unit AlgExp;
{ traitement des proprits algbriques }
{ KB dc 1997 }

INTERFACE

Uses AlgConst, AlgUtil, AlgNode, AlgSort, AlgCalc, AlgArit;

(* remplace soustraction et division par addition et multiplication
   x/y  :=  x*y^-1
   x-y  :=  x+y*-1
   renvoie 0 si aucune transformation possible *)
Function NoSubDiv(p:PNode):Integer;

(* Convertion des expressions dans la forme canonique o
   1. + et * sont associs  gauche, ^ est associ  droite.
   2. recherche des / avant * , e.g. (a/b)*c => ac/b,
   3. x+y*-2 est chang en x-y*2
   3' x+(-2) est chang en x-2
   4. x*y^-2 est chang en x/y^2
   5. rduction des ^
   renvoie 0 si aucune transformation possible *)
Function Bisect(p:PNode):Integer;

(* Remplace les puissances par des produits pour les exposants entiers
   infrieurs  100.
   renvoie 0 si aucune transformation possible *)
Function ExExpand(p:PNode):Integer;

(* si p est divisible par q, renvoie le quotient p/q.
   si le rsultat n'est pas nil, il s'agit d'une nouvelle copie. *)
Function Within(p,q:PNode):PNode;

(* Rduire au mme dnominateur  *)
Function ComDeno(p:PNode):Integer;

(* Simplify
   Cette fonction convertit une expression rationnelle  sa forme
   normale, c'est  dire :
   pas d'exposant ngatif, termes tris en fonction des multiplications
   et additions.
   La premire opration met tous les nombres  gauche, ainsi la
   fonction de tri les replace  droite et les combine.  *)
Procedure Simplify(p:PNode);

(* Simplification rduite *)
Procedure Simplif1(p:PNode);

(* Distribute
   Applique les proprits de distributivit
   1. multiplication sur addition.
   2. (ab)^2 := a^2*b^2
   3. x^(a+b) := x^a*x^b  *)
Function Distribute(p:PNode):Integer;

(* DistributeDiv
   Application de la distributivit de la division sur l'addition. *)
Function DistributeDiv(p:PNode):Integer;

(* DistributeFonc
   Distributivit des fonctions sur les oprations *)
Function DistributeFonc(p:PNode):Integer;

(* Substitution
  tgt est une galit var=exp;
  remplace var par exp dans p  *)
Procedure Substitution(p,tgt:PNode);

(* Combine
  Combine tous les termes de mme base. *)
Function Combine(p:PNode):Integer;

(* Fixassoc
   + et * sont associs  gauche, ^ est associ  droite.
   Rgles des parenthses *)
Function FixAssoc(p:PNode):Integer;

(* fournit un nouveau noeud avec le rsultat *)
Function AlgEval(p:PNode):PNode;
Procedure AlgEval1(p1:PNode);

(*
   associate
   rotate the association on add,sub,mul,div
   a+b-c => a-c+b => b+a-c => b-c+a
   note, we make use of the fact that _ADD+1 = _SUB and _ADD is even.
*)
Procedure Associate(p:PNode);

(* CrossEq
   transposition des termes d'une quation
   pour addition, soustraction et division c'est le 2me terme qui
   est transpos, pour multiplication c'est le premier
*)
Procedure CrossEq(p:PNode; i:Integer);

(* ExpJoin
   a^x * a^y => a^(x+y)
   x^a * y^a => (xy)^a
*)
Function Expjoin(p:PNode):Integer;

IMPLEMENTATION

(* DistribFonc
   Distributivit d'une fonction sur une opration *)
Function DistribFonc(p:PNode; nom:TName; oper:Integer):Integer;
Var r:Integer;
    a,tmp1,tmp2:PNode;
Begin
 r:=0;
 a:=p^.parm[0];
 if (p^.kind=_FUN) and (p^.name=nom)
    and (p^.nump=1) and (a^.kind=oper) then
  begin
   tmp1:=consfunc(nom,a^.parm[0]);
   tmp2:=consfunc(nom,a^.parm[1]);
   p^.kind:=a^.kind;
   p^.name:=a^.name;
   freenode(a);
   p^.nump:=2;
   p^.parm[0]:=tmp1;
   p^.parm[1]:=tmp2;
   inc(r);
  end;
 DistribFonc:=r;
End;

Procedure AlgEval1(p1:PNode);
Var ok1,ok2:Boolean;
Begin
 Simplify(p1);
 repeat
  ok1:=false; ok2:=false;
  (* distribuer les fonctions *)
  while DistributeFonc(p1)<>0 do ok1:=true;
  (* dcomposer les puissances en produits *)
  while ExExpand(p1)<>0 do ok1:=true;
  (* distribuer *)
  while Distribute(p1)<>0 do ok2:=true;
  (* si changements, rduire *)
  if ok1 or ok2 then Simplify(p1);
  (* si on n'a pas distribu on arrte *)
 until not ok2;
 repeat
  ok1:=false;
  while ComDeno(p1)<>0 do ok1:=true;
  if ok1 then
   begin
    while DistributeFonc(p1)<>0 do ;
    while Distribute(p1)<>0 do ;
    Simplify(p1);
   end;
 until not ok1;
End;

Function AlgEval(p:PNode):PNode;
Var p1:PNode;
Begin
 p1:=deepcopy(p);
 AlgEval1(p1);
 AlgEval:=p1;
End;

{ SIMPLIFY
  Convertit un noeud dans sa forme normale
}
Procedure Simplify(p:PNode);
Begin
 while calcnode(p)<>0 do ;
 while fixassoc(p)<>0 do ;
 while nosubdiv(p)<>0 do ;
 while fixassoc(p)<>0 do ;
 while movenums(p,0,_MUL)<>0 do ;
 if (sortnode(p,_MUL)<>0) then ;
 while movenums(p,0,_ADD)<>0 do ;
 if (sortnode(p,_ADD)<>0) then ;
 while combine(p)<>0 do
  begin
   while calcnode(p)<>0 do ;
   while fixassoc(p)<>0 do ;
   while nosubdiv(p)<>0 do ;
   while fixassoc(p)<>0 do ;
   while movenums(p,0,_MUL)<>0 do ;
   if (sortnode(p,_MUL)<>0) then ;
   while movenums(p,0,_ADD)<>0 do ;
   if (sortnode(p,_ADD)<>0) then ;
  end;
 while bisect(p)<>0 do ;
 while calcnode(p)<>0 do ;
 while movenums(p,0,_MUL)<>0 do ;
End;

Procedure Simplif1(p:PNode);
Begin
 while calcnode(p)<>0 do ;
 while fixassoc(p)<>0 do ;
 {while nosubdiv(p)<>0 do ;
 while fixassoc(p)<>0 do ;}
 if (sortnode(p,_MUL)<>0) then ;
 if (sortnode(p,_ADD)<>0) then ;
 while bisect(p)<>0 do ;
 while calcnode(p)<>0 do ;
 while movenums(p,0,_MUL)<>0 do ;
End;

(* NoSubDiv
   remplace soustraction et division par addition et multiplication
   x/y  :=  x*y^-1
   x-y  :=  x+y*-1
   x/rac(y) := x*rac(y)/y
*)
Function NoSubDiv(p:PNode):Integer;
Var i,r : Integer;
    a,b,tmp : PNode;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+NoSubdiv(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 if (p^.kind=_SUB) or (p^.kind=_DIV) then
  begin
   inc(r);         (* on effectue la transformation *)
   dec(p^.kind);   (* nouvelle opration : - vers + ou / vers * *)
   if (b^.kind=_NUM) and (p^.kind=_ADD) then   (* a-b := a+(-b) *)
    b^.value:=-b^.value  (* b est une valeur numrique *)
   else
    begin
     if (p^.kind=_MUL) and (b^.kind=_FUN) and (b^.name='rac') then
      begin     (*  a/rac(b) = a*rac(b)/b  *)
       tmp:=deepcopy(b^.parm[0]);
       p^.kind:=_DIV;
       p^.parm[0]:=cons(a,_MUL,b);
       p^.parm[1]:=tmp;
      end
     else
      begin
       a:=newoper(p^.kind+2);  (* _MUL ou _EXP *)
       a^.parm[0]:=b;
       a^.parm[1]:=newnum(-1);
       p^.parm[1]:=a;
      end;
    end;
  end;
 NoSubdiv:=r;
End;

(* Bisect
   Convertit une expression dans sa forme canonique *)
Function Bisect(p:PNode):Integer;
Var i,r : Integer;
    a,b,tmp : PNode;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+Bisect(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 case p^.kind of
  _ADD, _MUL: (* addition et multiplication *)
   begin
    if (b^.kind=p^.kind) or (b^.kind=p^.kind+1) then
     begin  (* u+(v-w)=u+v-w ou avec *  et  u+(v+w)=u+v+w ou avec *  *)
      swingb(a,b,p);             (* change l'arbre : b remplace a *)
      i:=b^.kind;                (* avec mise  jour coeffs *)
      b^.kind := p^.kind;        (* mise  jour opers *)
      p^.kind := i;
     end
    else if (p^.kind=_ADD) and (b^.kind=_NUM) and (b^.value < 0) then
     begin  (*  u+(-2)=u-2  *)
      p^.kind:=_SUB;
      b^.value:=-b^.value;
     end
    else if (b^.kind=p^.kind+2) and (b^.parm[1]^.kind=_NUM)
          and (b^.parm[1]^.value<0) then
     begin (*  u+v*-2=u-v*2  *)
      inc(p^.kind);
      b^.parm[1]^.value:=-b^.parm[1]^.value;
     end
    else if (p^.kind=_ADD) and aop(b^.kind) then
     begin   (*  a+(b+c) := (a+b)+c  *)
      swingb(a,b,p);
      p^.kind:=b^.kind;
      b^.kind:=_ADD;
     end
    else if (p^.kind=_MUL) and (a^.kind=p^.kind+1) then
     begin  (*  a/b*c = a*c/b  *)
      p^.parm[1]:=a^.parm[1];
      a^.parm[1]:=b;
      inc(p^.kind);
      dec(a^.kind);
     end
    else if (p^.kind=_ADD) and (a^.kind=_MUL) and (a^.parm[1]^.kind=_NUM)
          and (a^.parm[1]^.value < 0) then
     begin   (*  b*-2+a = a-b*2  *)
      inc(p^.kind);
      p^.parm[0]:=b;
      p^.parm[1]:=a;
      a^.parm[1]^.value:=-a^.parm[1]^.value;
     end
    else dec(r);
    inc(r);
   end;
  _SUB, _DIV: (* soustraction et division *)
   begin
    if (b^.kind=p^.kind+1) and (b^.parm[1]^.kind=_NUM) and
       (b^.parm[1]^.value < 0) then
     begin  (* a-b*-2 := a+b*2  non ncessaire *)
      dec(p^.kind);
      b^.parm[1]^.value:=-b^.parm[1]^.value;
     end
    else if (p^.kind=_SUB) and (b^.kind=_NUM) and (b^.value < 0) then
     begin   (* a-(-2) := a+2 *)
      p^.kind:=_ADD;
      b^.value:=-b^.value;
     end
    else if (p^.kind=_SUB) and (aop(b^.kind)) then
     begin  (* a-(b+c) := (a-b)-c *)
      swingb(a,b,p);
      p^.kind:=(_ADD+_SUB)-b^.kind;
      b^.kind:=_SUB;
     end
    else if (p^.kind=_DIV) and (b^.kind=p^.kind) then
     begin  (* a/(b/c) := a*c/b *)
      p^.parm[0] := b;
      p^.parm[1] := b^.parm[0];
      b^.parm[0] := a;
      dec(b^.kind);
     end
    else if (p^.kind=_DIV) and (a^.kind=p^.kind) then
     begin  (* a-b-c := a-(b+c) *)
      swinga(a,b,p);
      dec(a^.kind);
     end
    else dec(r); (* pas de transformation *)
    inc(r);
   end;
  _EXP:  (* puissances *)
   begin
    if (a^.kind=_EXP) then
     begin    (* (x^y)^z := x^(y*z) *)
      swinga(a,b,p);
      a^.kind:=_MUL;
     end
    else if (b^.kind=_NUM) and (b^.value<0) then
     begin    (* a^(-2) := 1/a^2 *)
      p^.kind:=_DIV;
      p^.parm[0]:=newnum(1);
      p^.parm[1] := newoper(_EXP);
      p^.parm[1]^.parm[0] := a;
      p^.parm[1]^.parm[1] := b;
      b^.value:=-b^.value;
     end
    else dec(r); (* pas de transformation *)
    inc(r);
   end;
  end;
 Bisect:=r;
End;

(* ExExpand
   Remplace les puissances  exposants entiers (jusqu' 100) par des
   produits.
*)
Function ExExpand(p:PNode):Integer;
Var i,r:Integer;
    a,b:PNode;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+ExExpand(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 if (p^.kind=_EXP) and (b^.kind=_NUM) and (b^.value>1)
   and (b^.value<=100)
   and (a^.kind<>_NUM) then
  begin   (* a^n := a^(n-1)*a *)
   b^.value:=b^.value-1;
   if b^.value=1 then
    begin
     p^.parm[0]:=deepcopy(a);
     freenode(b);
    end
   else
    begin
     b:=newoper(_EXP);
     b^.parm[0]:=deepcopy(a);
     b^.parm[1]:=p^.parm[1];
     p^.parm[0]:=b;
    end;
   p^.parm[1]:=a;
   p^.kind:=_MUL;
   inc(r);
  end;
 ExExpand:=r;
End;

(* Within
   Si p a q comme facteur, renvoie le reste q.
   Si le rsultat n'est pas nil, il s'agit d'une nouvelle copie.
*)
Function Within(p,q:PNode):PNode;
Var a,b:PNode;
Begin
 a:=q; b:=nil;
 while a^.kind=_MUL do
  begin
   if equal(p,a^.parm[1]) then
    begin
     if b<>nil then
      begin
       b^.parm[0]:=a^.parm[0];    (* skip over match *)
       q:=deepcopy(q);            (* make copy *)
       b^.parm[0]:=a;             (* repair *)
      end
     else q:=deepcopy(a^.parm[0]);
     Within:=q;
     exit;
    end
   else if equal(p,a^.parm[0]) then
    begin
     if b<>nil then
      begin
       b^.parm[0]:=a^.parm[1];   (* skip over match *)
       q:=deepcopy(q);           (* make copy *)
       b^.parm[0]:=a;            (* repair *)
      end
     else q:=deepcopy(a^.parm[1]);
     Within:=q;
     exit;
    end;
   b:=a;
   a:=b^.parm[0];
  end;
 Within:=nil;
End;

(* ComDeno
   Rduction au mme dnominateur.
*)
Function ComDeno(p:PNode):Integer;
Var i,r:Integer;
    x:TDouble;
    a,b,nu,de:PNode;
    l:LongInt;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+comdeno(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 if aop(p^.kind) and ((a^.kind=_DIV) or (b^.kind=_DIV)) then
  begin
   if (a^.kind=_DIV) and (b^.kind=_DIV) then
    begin  (* somme de quotients *)
     if equal(a^.parm[1],b^.parm[1]) then
      begin     (* u/x + v/x := (u+v)/x *)
       freetree(b^.parm[1]);
       de:=a^.parm[1];
       a:=a^.parm[0];
       b:=b^.parm[0];
       freenode(p^.parm[0]);
       freenode(p^.parm[1]);
      end
     else
      begin
       nu:=within(a^.parm[1],b^.parm[1]);
       if nu<>nil then
        begin   (* u/x + v/xy := (uy+v)/xy *)
         freetree(a^.parm[1]);
         a:=cons(nu,_MUL,a^.parm[0]);
         de:=b^.parm[1];
         b:=b^.parm[0];
         freenode(p^.parm[0]);
         freenode(p^.parm[1]);
        end
       else
        begin
         nu:=within(b^.parm[1],a^.parm[1]);
         if nu<>nil then
          begin   (* u/xy + v/x := (u+vy)/xy *)
           freetree(b^.parm[1]);
           b:=cons(nu,_MUL,b^.parm[0]);
           de:=a^.parm[1];
           a:=a^.parm[0];
           freenode(p^.parm[0]);
           freenode(p^.parm[1]);
          end
         else if (a^.parm[1]^.kind=_NUM) and (b^.parm[1]^.kind=_NUM) then
          begin  (* u/c + v/d := (ud+vc)/cd  avec c et d entiers *)
           l:=PPCM(a^.parm[1]^.value,b^.parm[1]^.value);
           a^.kind:=_MUL;
           b^.kind:=_MUL;
           a^.parm[1]^.value:=l div a^.parm[1]^.value;
           b^.parm[1]^.value:=l div b^.parm[1]^.value;
           de:=newnum(l);
          end
         else
          begin
           a^.kind:=_MUL;
           b^.kind:=_MUL;
           de:=b^.parm[1];
           b^.parm[1]:=a^.parm[1];
           a^.parm[1]:=de;
           de:=newoper(_MUL);
           de^.parm[0]:=deepcopy(b^.parm[1]);
           de^.parm[1]:=deepcopy(a^.parm[1]);
          end;
         end;
      end;
     end (* fin somme de quotients *)
    else if a^.kind=_DIV then
     begin   (* a/b + c := (a + bc)/b *)
      a^.kind:=_MUL;
      de:=a^.parm[0];
      a^.parm[0]:=b;
      b:=a;
      a:=de;
      de:=deepcopy(b^.parm[1]);
     end
    else
     begin  (* a + b/c := (ac + b)/c *)
      b^.kind:=_MUL;
      de:=b^.parm[0];
      b^.parm[0]:=a;
      a:=b;
      b:=de;
      de:=deepcopy(a^.parm[1]);
     end;
   nu:=newoper(p^.kind);
   nu^.parm[0]:=a;
   nu^.parm[1]:=b;
   p^.kind:=_DIV;
   p^.parm[0]:=nu;
   p^.parm[1]:=de;
   inc(r);
  end;
 ComDeno:=r;
End;

(*
   DistributeDiv
   Application de la distributivit de la division sur l'addition.  *)
Function DistributeDiv(p:PNode):Integer;
Var a,b,c : PNode;
    i,r : Integer;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+DistributeDiv(p^.parm[i]);
 i:=0;
 c := p^.parm[i];
 if (p^.kind=_DIV) and aop(c^.kind) then
  begin
   b:=deepcopy(p^.parm[1-i]);
   a:=newoper(p^.kind);
   a^.parm[i]:=c^.parm[1-i];
   a^.parm[1-i]:=b;
   c^.parm[1-i]:=p^.parm[1-i];
   p^.parm[1-i]:=a;
   p^.kind:=c^.kind;
   c^.kind:=a^.kind;
   inc(r);
  end;
 DistributeDiv:=r;
End;

(* Constante utilise pour dterminer la direction de la distributivit *)
Const topdown:boolean=false;

(* Distribute
   Applique la distributivit :
   1. multiplication sur addition.
   2. (ab)^2 := a^2*b^2
   3. x^(a+b) := x^a*x^b
*)
Function Distribute(p:PNode):Integer;
Var a,b,c:PNode;
    i,r:Integer;
Begin
 r:=0;
 if not topdown
  then for i:=0 to p^.nump-1 do r:=r+distribute(p^.parm[i]);
 i:=0;
 while i<2 do
  begin
   c:=p^.parm[i];
   if ((p^.kind=_MUL) and aop(c^.kind))
    or ((i<>0) and (p^.kind=_EXP) and aop(c^.kind))
    or ((i=0) and (p^.kind=_EXP) and ((c^.kind=_MUL)
        or (c^.kind=_DIV))) then
    begin
     b:=deepcopy(p^.parm[1-i]);
     a:=newoper(p^.kind);
     a^.parm[i]:=c^.parm[1-i];
     a^.parm[1-i]:=b;
     c^.parm[1-i]:=p^.parm[1-i];
     p^.parm[1-i]:=a;
     p^.kind:=c^.kind;
     c^.kind:=a^.kind;
     if (i<>0) and (a^.kind=_EXP) then
      inc(p^.kind,2);    (* changement de _ADD en _MUL *)
     inc(r);
     i:=1;   (* juste pour tre sr *)
    end;
   inc(i);
  end;
 if topdown and (r=0)
  then for i:=0 to p^.nump-1 do r:=r+distribute(p^.parm[i]);
 Distribute:=r;
End;

(* DistributeFonc
   Distributivit des fonctions sur les oprations *)
Function DistributeFonc(p:PNode):Integer;
Var r,i:Integer;
    a,b:PNode;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+DistributeFonc(p^.parm[i]);
 if (p^.kind<>_FUN) then
  begin
   DistributeFonc:=0;
   exit;
  end;
 if p^.name='rac' then
  begin    (*  rac(x*y) = rac(abs(x)) * rac(abs(y))   *)
   if (p^.parm[0]^.kind=_MUL) or (p^.parm[0]^.kind=_DIV) then
    begin
     if p^.parm[0]^.parm[0]^.kind=_NUM then
      begin
       p^.parm[0]^.parm[0]^.value:=abs(p^.parm[0]^.parm[0]^.value);
       a:=p^.parm[0]^.parm[0];
      end
     else a:=consfunc('abs',p^.parm[0]^.parm[0]);
     if p^.parm[0]^.parm[1]^.kind=_NUM then
      begin
       p^.parm[0]^.parm[1]^.value:=abs(p^.parm[0]^.parm[1]^.value);
       b:=p^.parm[0]^.parm[1];
      end
     else b:=consfunc('abs',p^.parm[0]^.parm[1]);
     p^.parm[0]^.parm[0]:=a;
     p^.parm[0]^.parm[1]:=b;
     case p^.parm[0]^.kind of
      _MUL: r:=r+DistribFonc(p,'rac',_MUL);
      _DIV: r:=r+DistribFonc(p,'rac',_DIV);
      end
    end;
  end
 else if p^.name='abs' then
  begin   (*  abs(x*y) = abs(x) * abs(y)  *)
   case p^.parm[0]^.kind of
    _MUL: r:=r+DistribFonc(p,'abs',_MUL);
    _DIV: r:=r+DistribFonc(p,'abs',_DIV);
    end;
  end;
 DistributeFonc:=r;
End;

(* Fixassoc
   + et * sont associs  gauche, ^ est associ  droite.
*)
Function Fixassoc(p:PNode):Integer;
Var
 i,r:Integer;
 a,b:PNode;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+FixAssoc(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 case p^.kind of
  _ADD,_MUL:
   begin
    if (b^.kind=p^.kind) or              (* a+(b+c) := a+b+c *)
       (b^.kind=p^.kind+1) then          (* a+(b-c) := a+b-c *)
     begin
      swingb(a,b,p);
      i:=b^.kind; b^.kind:=p^.kind; p^.kind:=i;
     end
    else dec(r);
    inc(r);
   end;
  _SUB,_DIV:
   begin
    if b^.kind=p^.kind then
     begin             (* a-(b-c) := a-b+c *)
      swingb(a,b,p);
      dec(p^.kind);
     end
    else if b^.kind=p^.kind-1 then
     begin      (* a-(b+c) := a-b-c *)
      swingb(a,b,p);
      inc(b^.kind);
     end
    else dec(r);
    inc(r);
   end;
  _EXP:
   begin
    if a^.kind=_EXP then
     begin                  (* (x^y)^z := x^(y*z) *)
      swinga(a,b,p);
      a^.kind:=_MUL;
     end
    else dec(r);
    inc(r);
   end;
  end;
 Fixassoc:=r;
End;

(* GetTerm
   renvoie un terme, et la valeur de son coefficient
*)
Function GetTerm(p:PNode; oper:Integer; Var r:TDouble):PNode;
Begin
 r:=1;              (* coefficient par dfaut *)
 if (p^.kind=oper) and (p^.parm[1]^.kind=_NUM) then
  begin
   r:=p^.parm[1]^.value;
   getterm:=p^.parm[0]
  end
 else getterm:=p;
End;

(* Combine
   Combine les termes de mme base.
*)
Function Combine(p:PNode):Integer;
Var i,oper,r:Integer;
    a,b,t1,t2:PNode;
    c1,c2:TDouble;
Begin
 r:=0;
 for i:=0 to p^.nump-1 do r:=r+combine(p^.parm[i]);
 a:=p^.parm[0];
 b:=p^.parm[1];
 oper:=p^.kind;
 if (oper<>_ADD) and (oper<>_MUL) then
  begin
   Combine:=r;
   exit;
  end;
 if (oper=_MUL) and (a^.kind=_FUN) and (a^.name='rac')
    and (b^.kind=_FUN) and (b^.name='rac') then
  begin   (* produit de racines *)
   a:=a^.parm[0];
   freenode(p^.parm[0]);
   b:=b^.parm[0];
   freenode(p^.parm[1]);
   t1:=cons(a,oper,b);
   p^.kind:=_FUN;
   p^.name:='rac';
   p^.nump:=1;
   p^.parm[0]:=t1;
   p^.parm[1]:=nil;
   Combine:=r+1;
   exit;
  end;
 i:=1;
 if a^.kind<>oper then
  begin
   a:=p; i:=0;
  end;
 t1:=GetTerm(a^.parm[i],oper+2,c1);
 t2:=GetTerm(b,oper+2,c2);
 if equal(t1,t2) then
  begin            (* termes  combiner *)
   if a^.parm[i]=t1 then
    freetree(t1)
   else
    begin           (* free one of the expressions *)
     freetree(b);
     b:=a^.parm[i];
     t2:=t1;
    end;
   c2:=c2+c1;
   if b=t2 then
    begin             (* make room for coeff *)
     b:=newoper(oper+2);
     b^.parm[0]:=t2;
     b^.parm[1]:=newnum(0);
    end;
   b^.parm[1]^.value := c2;
   if c2=1 then
    begin            (* remove unit coeff *)
     freenode(b^.parm[1]);
     freenode(b);
     b:=t2;
    end;
   if i<>0 then
    begin                  (* cleanup... 2 stage *)
     nodecpy(p,a);
     freenode(a);
     p^.parm[1] := b;
    end
   else
    begin                    (* 1 stage *)
     nodecpy(p,b);
     freenode(b);
    end;
   inc(r);
  end;
 Combine:=r;
End;

(* Substitute
   Remplacement d'une variable par une expression
   tgt DOIT tre une galit du type " nomvar = expr ".
*)
Procedure Substitution(p,tgt:PNode);
Var i:Integer;
Begin
 if tgt=nil then exit;
 if equal(p,tgt^.parm[0])
  then movenode(p,deepcopy(tgt^.parm[1]))
  else for i:=0 to p^.nump-1 do substitution(p^.parm[i],tgt);
End;

(* Associate
   a+b-c => a-c+b => b+a-c => b-c+a
   Note : on utilise le fait que _ADD+1=_SUB et _ADD est pair.
*)
Procedure Associate(p:PNode);
Var a,b : PNode;
    opr : Integer;
Begin
 opr:=p^.kind;
 b:=p^.parm[1];
 if (opr=_DIV) and (b^.kind=_MUL) then
  begin     (* traitement spcial pour fonction Bisect *)
   p^.parm[1]:=b^.parm[1];
   b^.parm[1]:= b^.parm[0];
   b^.parm[0]:=p^.parm[0];
   p^.parm[0]:=b;
   b^.kind:=_DIV;
   b:=p^.parm[1];
  end;
 if (opr=_ADD) or (opr=_MUL) or (opr=_SUB) or (opr=_DIV) then
  begin
   a:=p;
   while ((a^.parm[0]^.kind or 1)=(opr or 1)) do
    begin
     a^.kind:=a^.parm[0]^.kind;
     a^.parm[1]:=a^.parm[0]^.parm[1];
     a:=a^.parm[0];
    end;
   a^.kind:=opr;
   if opr and 1 <> 0 then
    begin           (* soustraction ou division *)
     a^.parm[1]:=b;
    end
   else
    begin
     a^.parm[1]:=a^.parm[0];
     a^.parm[0]:=b;
    end;
  end
  else if opr=_EQU then   (* galit *)
   begin    (* commute les 2 membres *)
    p^.parm[1]:=p^.parm[0];
    p^.parm[0]:=b;
   end;
End;

(*
   invert left side of equality, i:=0 or 1
*)
Procedure CrossEq(p:PNode; i:Integer);
Var g,d,t:PNode;
    k:Integer;
Begin
 if p^.kind<>_EQU then exit;
 g:=p^.parm[0];
 d:=p^.parm[1];
 case i of
  0 :
   begin
    case g^.kind of
     _ADD,_SUB,_DIV :
      begin
       p^.parm[1]:=cons(d,g^.kind xor 1,g^.parm[1]);
       t:=g^.parm[0];
       freenode(g);
       p^.parm[0]:=t;
      end;
     _MUL :
      begin
       p^.parm[1]:=cons(d,g^.kind xor 1,g^.parm[0]);
       t:=g^.parm[1];
       freenode(g);
       p^.parm[0]:=t;
      end;
     else
      begin
       p^.parm[1]:=cons(d,_SUB,deepcopy(g));
       for k:=0 to g^.nump-1 do freetree(g^.parm[k]);
       g^.kind:=_NUM;
       g^.nump:=0;
       g^.value:=0;
      end;
    end;
   end;
  1 :
   begin
    case d^.kind of
     _ADD,_SUB,_DIV :
      begin
       p^.parm[0]:=cons(g,d^.kind xor 1,d^.parm[1]);
       t:=d^.parm[0];
       freenode(d);
       p^.parm[1]:=t;
      end;
     _MUL :
      begin
       p^.parm[0]:=cons(g,d^.kind xor 1,d^.parm[0]);
       t:=d^.parm[1];
       freenode(d);
       p^.parm[1]:=t;
      end;
     else
      begin
       p^.parm[0]:=cons(g,_SUB,deepcopy(d));
       for k:=0 to d^.nump-1 do freetree(d^.parm[k]);
       d^.kind:=_NUM;
       d^.nump:=0;
       d^.value:=0;
      end;
    end;
   end;
 end;
End;

(* ExpJoin
      a^x * a^y => a^(x+y)
      rac(a) * rac(b) => rac(a*b)
*)
Function Expjoin(p:PNode):Integer;
Var i,r:Integer;
    a,b,tmp:PNode;
Begin
 (*  rponse par dfaut : nombre de transformations ralises  *)
 r:=0;
 (*  traitement des paramtres  *)
 for i:=0 to p^.nump-1 do r:=r+ExpJoin(p^.parm[i]);
 (*  a et b sont les deux termes de l'opration  *)
 a:=p^.parm[0];
 b:=p^.parm[1];
 if (p^.kind=_MUL) or (p^.kind=_DIV) then
  begin  (*  multiplication et division  *)
   if (a^.kind=_EXP) and (b^.kind=_EXP) and
      equal(a^.parm[0],b^.parm[0]) then
    begin      (*  u^x * u^y := u^(x+y)  *)
               (*  on va de a=u^x et b=u^y et op=*(ou/)
                 vers a=u et b=x+(ou-)y et op=^  *)
     freetree(a^.parm[0]);    (*  dtruire u de u^x  *)
     a:=a^.parm[1];           (*  a=x  *)
     freenode(p^.parm[0]);    (*  librer a  *)
     b^.kind:=p^.kind-2;      (*  opration associe  b : + ou -  *)
     p^.kind:=_EXP;           (*  opration associe  p : ^  *)
     p^.parm[0]:=b^.parm[0];  (*  nouveau a=u  *)
     b^.parm[0]:=a;           (*  rcupre x pour b  *)
    end
   else if (a^.kind=_FUN) and (a^.name='rac')
       and (b^.kind=_FUN) and (b^.name='rac') then
    begin (* produit de racines *)
     a:=a^.parm[0];
     freenode(p^.parm[0]);
     b:=b^.parm[0];
     freenode(p^.parm[1]);
     tmp:=cons(a,p^.kind,b);
     p^.kind:=_FUN;
     p^.name:='rac';
     p^.nump:=1;
     p^.parm[0]:=tmp;
     p^.parm[1]:=nil;
    end
   else if (b^.kind=_EXP) and equal(a,b^.parm[0]) then
    begin         (*  u * u^y := u^(1+y)  *)
     freetree(a);
     a:=newnum(1);
     b^.kind:=p^.kind-2;
     p^.kind:=_EXP;
     p^.parm[0]:=b^.parm[0];
     b^.parm[0]:=a;
    end
   else if (a^.kind=_EXP) and equal(a^.parm[0],b) then
    begin      (*  u^x * u := u^(x+1)  *)
     freetree(b);
     b:=newnum(1);
     a^.kind:=p^.kind-2;
     p^.kind:=_EXP;
     p^.parm[0]:=a^.parm[0];
     a^.parm[0]:=a^.parm[1];
     a^.parm[1]:=b;
     p^.parm[1]:=a;
    end
   else dec(r);
  end
 else if (p^.kind=_EXP) and (p^.parm[0]^.kind=_EXP) then
  begin  (*  (a^x)^y = a^(x*y)    *)
   swinga(a,b,p);
   a^.kind:=_MUL;
  end
 else dec(r);
 inc(r);
 ExpJoin:=r;
End;

END.

{                         Fin du fichier ALGEXP.PAS                         }
