Unit AlgNode;
{ gestion des noeuds dcrivant une expression algbrique }
{ KB dc 1997 }

INTERFACE

Uses AlgConst, AlgUtil;

Const
 MAXP=5;   { nombre maximal de paramtres pour un noeud }

Type
{ Structure fondamentale pour les formules }

 PNode = ^TNode;
 TNode = record
  name  : TName;                     { libell }
  parm  : array[0..MAXP-1] of PNode; { paramtres }
  tag   : Integer;                   { marqueur normalement gal  12345 }
  value : TDouble;                   { valeur }
  sx    : Integer;                   { taille  l'cran en largeur }
  px,py : Integer;                   { position  l'cran }
  ay    : Integer;                   { position de base }
  dyh,dyb : Integer;                 { dbordements haut et bas }
  kind  : Byte;                      { type de noeud : voir plus bas }
  mpr   : Byte;                      { niveau de parenthses }
  nump  : Integer;                   { nombre de paramtres }
  Msg   : PStr80;                    { pointeur sur message }
  next  : PNode;                     { noeud suivant pour liste lie }
  end;

Function NewNode:PNode;
{ cration d'un noeud avec valeurs par dfaut }

Function NewNum(val:TDouble):PNode;
{ cration d'un noeud reprsentant une valeur numrique }

Function NewReal(r:Real):PNode;
{ cration d'un noeud reprsentant un nombre dcimal }

Function NewVar(s:TName):PNode;
{ cration d'un noeud reprsentant une variable }

Function NewOper(kind:Integer):PNode;
{ cration d'un noeud reprsentant une opration }

Procedure FreeNode(p:PNode);
{ libration de la mmoire occupe par un noeud }

Procedure FreeTree(p:PNode);
{ libration de la mmoire occupe par un noeud ainsi que ses paramtres }

Procedure Nodecpy(a,b:PNode);
{ copie du noeud b dans le noeud a, le suivant de a dans la liste lie
  est prserv }

Procedure MoveNode(a,b:PNode);
{ dtruit a et copie b dans a }

Function Cons(left:PNode;opr:Integer;right:PNode):PNode;
{ cration d'un noeud opration avec ses deux membres }

Function ConsFunc(name:TName; par:PNode):PNode;
{ cration d'un noeud fonction avec sa variable unique }

Function DeepCopy(p:PNode):PNode;
{ cration d'un noeud identique  p }

Function LastNode(p:PNode):PNode;
{ dernier noeud dans la liste lie }

Function Reverse(b:PNode):PNode;
{ renverse l'ordre des noeuds de la liste lie de b jusqu' la fin }

Procedure swingb(a,b,p:PNode);
Procedure swinga(a,b,p:PNode);

Procedure SetNumName(p:PNode);

Function GetNodeStr(p:PNode; ppr:Integer):String;
{ renvoie une chane correspondant au noeud }

Function Evaluable(p:PNode):Boolean;

IMPLEMENTATION

Function Evaluable(p:PNode):Boolean;
Begin
 Evaluable:=(p<>nil) and (p^.kind<=_DEC);
End;

Procedure swingb(a,b,p:PNode);
Begin
 p^.parm[1]:= b^.parm[1];
 b^.parm[1]:= b^.parm[0];
 b^.parm[0]:= p^.parm[0];
 p^.parm[0]:= b;
End;

Procedure swinga(a,b,p:PNode);
Begin
 p^.parm[0] := a^.parm[0];
 a^.parm[0] := a^.parm[1];
 a^.parm[1] := p^.parm[1];
 p^.parm[1] := a;
End;

Function NewNode:PNode;
Var p : PNode;
    i : Integer;
Begin
 New(p);
 if p<>nil then
 with p^ do
  begin
   tag:=12345;
   nump:=0;
   msg:=nil;
   kind:=_VAR;
   name:='!';
   px:=0; py:=0;
   sx:=0;
   ay:=0;
   dyh:=0; dyb:=0;
   mpr:=0;
   for i:=0 to MAXP-1 do parm[i]:=nil;
  end;
 NewNode:=p;
End;

Procedure SetNumName(p:PNode);
Var x:TDouble;
Begin
 x:=abs(p^.value);
 p^.name:='';
 p^.name:=IntToStr(x);
 if p^.value<0 then p^.name:='(-'+p^.name+')';
End;

Function NewNum(val:TDouble):PNode;
Var p : PNode;
Begin
 p:=NewNode;
 if p<>nil then
  begin
   p^.kind:=_NUM;
   p^.value:=val;
   SetNumName(p);
  end;
 NewNum:=p;
End;

Function NewReal(r:Real):PNode;
Var p:PNode;
    S:String[40];
Begin
 p:=NewNode;
 if p<>nil then
  begin
   p^.kind:=_DEC;
   p^.value:=0;
   str(r:0:8,S);
   p^.name:=S;
  end;
 NewReal:=p;
End;

Function NewVar(s:TName):PNode;
Var p : PNode;
Begin
 p:=newnode;
 if p<>nil then p^.name:=s;
 NewVar:=p;
End;

Function NewOper(kind:Integer):PNode;
Var p : PNode;
Begin
 p:=NewNode;
 if p<>nil then
  begin
   p^.nump:=2;
   p^.kind:=kind;
   p^.name:=kname[kind];
  end;
 NewOper:=p;
End;

Procedure FreeNode(p:PNode);
Begin
 if (p=nil) or (p^.tag <> 12345) then
  begin
   writeln('Erreur pointeur.');
   exit;
  end;
 p^.tag:=0;
 p^.nump:=0;
 if p^.msg<>nil then freemem(p^.msg,length(p^.msg^)+1);
 p^.msg:=nil;
 p^.kind:= _VAR;
 p^.name:='!';
 dispose(p);
End;

Procedure FreeTree(p:PNode);
Var i,n : Integer;
Begin
 if (p=nil) or (p^.tag<>12345) then
  begin
   writeln('Erreur pointeur.');
   exit;
  end;
 n:=p^.nump;
 p^.tag:=0;
 p^.nump:=0;
 if p^.msg<>nil then freemem(p^.msg,length(p^.msg^)+1);
 p^.msg:=nil;
 p^.kind:=_VAR;
 p^.name:='';
 for i:=0 to n-1 do if p^.parm[i]<>nil then freetree(p^.parm[i]);
 dispose(p);
End;

(*
   nodecpy, use this to copy one node over another.  The only
   thing is that "next" is preserved to not disrupt the main
   node linked list.
*)
Procedure Nodecpy(a,b:PNode);
Var nx:PNode;
    x,y:Integer;
Begin
 x:=a^.px; y:=a^.py;
 nx:=a^.next;
 a^:=b^;
 a^.next:=nx;
 a^.px:=x; a^.py:=y;
End;

(*
   set a=b,
   frees a, then copies b into it and frees b.
*)
Procedure MoveNode(a,b:PNode);
Var i:Integer;
Begin
 (* destruction des enfants *)
 for i:=0 to a^.nump-1 do
  if a^.parm[i]<>nil then freetree(a^.parm[i]);
 nodecpy(a,b);
 freenode(b);
End;

(*
   cons  -- create a new node with left,opr,right or if left is
        null, just return right.
*)
Function Cons(left:PNode;opr:Integer;right:PNode):PNode;
Var a:PNode;
Begin
 if left=nil
  then begin
        cons:=right;       (* ... *)
        exit;
       end;
 a:=newoper(opr);
 a^.parm[0]:=left;
 a^.parm[1]:=right;
 cons:=a;
End;

Function ConsFunc(name:TName; par:PNode):PNode;
Var a:PNode;
Begin
 a:=newnode;
 if a<>nil then
  begin
   a^.name:=name;
   a^.kind:=_FUN;
   a^.nump:=1;
   a^.parm[0]:=par;
  end;
 ConsFunc:=a;
End;

(*
   deepcopy a node
*)
Function Deepcopy(p:PNode):PNode;
Var
 j : Integer;
 t : PNode;
Begin
 t:=newnode;
 nodecpy(t,p);
 for j:=0 to p^.nump-1 do t^.parm[j]:= deepcopy(p^.parm[j]);
 deepcopy:=t;
End;

(*
   lastnode
*)
Function LastNode(p:PNode):PNode;
Begin
 while (p<>nil) and (p^.next<>nil) do p:=p^.next;
 lastnode:=p;
End;


(*
   reverse the node list
*)
Function Reverse(b:PNode):PNode;
{ renverse l'ordre des noeuds de la liste lie de b jusqu' la fin }
Var
 a,c:PNode;
Begin
 a:=nil;
 while b<>nil do
  begin
   c:=b^.next;
   b^.next:=a;
   a:=b;
   b:=c;
  end;
 reverse:=a;
End;

Function GetNodeStr(p:PNode; ppr:Integer):String;
Var S:TStr80;
    i,mpr:Integer;
Begin
 mpr:=rpr[p^.Kind];
 if p^.Kind=_NUM then SetNumName(p);
 case p^.Kind of
  _NUM,_BAD,_VAR,_DEC: GetNodeStr:=p^.name;
  _EQU,_ADD,_SUB,_MUL,_DIV,_EXP,_ENV:
   begin
    S:='';
    if mpr<ppr
     then S:='(';
    S:=S+GetNodeStr(p^.parm[0],mpr);
    if p^.Kind=_MUL
     then S:=S+'*'
     else S:=S+kname[p^.kind];
    S:=S+GetNodeStr(p^.parm[1],mpr+1);
    if mpr<ppr
     then S:=S+')';
    GetNodeStr:=S;
   end;
  _FUN:
   begin
    S:='';
    S:=S+p^.name+'(';
    i:=0;
    repeat
     S:=S+GetNodeStr(p^.parm[i],mpr);
     inc(i);
     if i < p^.nump
      then S:=S+',';
    until (i>p^.nump-1);
    S:=S+')';
    GetNodeStr:=S
   end;
  else GetNodeStr:='bizarre';
 end;
End;

END.

{                       Fin du fichier ALGNODE.PAS                          }
