(*--------------------------------------------------------------------
   Alged:  Algebra Editor    henckel@vnet.ibm.com

   Copyright (c) 1994 John Henckel
   Permission to use, copy, modify, distribute and sell this software
   and its documentation for any purpose is hereby granted without fee,
   provided that the above copyright notice appear in all copies.
*)
Unit AlgSort;

INTERFACE

Uses AlgConst, AlgUtil, AlgNode;

Function equal(p,q:PNode):Boolean;
Function cmp(p,q:PNode):Integer;
Function sortnode(p:PNode; oper:Integer):Integer;

IMPLEMENTATION

(*
   compare two nodes, if they are equal return 1, else 0.
*)
Function Equal(p,q:PNode):Boolean;
Var
 i:Integer;
Begin
 if (p^.kind <> q^.kind) or (p^.nump <> q^.nump) then
  begin  { types diffrents ou nombres de paramtres diffrents }
   equal:=false;
   exit;
  end;
 if p^.kind=_NUM then
  begin  { valeurs numriques : mme valeur ? }
   equal:= (p^.value=q^.value);
   exit;
  end;
 if p^.kind=_VAR then
  begin  { variables : mme nom ? }
   equal:= (p^.name=q^.name);
   exit;
  end;
 if (p^.kind=_FUN) and (p^.name<>q^.name) then
  begin  { fonctions de noms diffrents }
   equal:=false;
   exit;
  end;
 for i:=0 to p^.nump-1 do  { autres cas => comparaison des paramtres }
  if not equal(p^.parm[i],q^.parm[i]) then
   begin
    equal:=false;
    exit;
   end;
 equal:=true;             (* identical *)
End;

(*-----------------------------------------------------------------
   compare two nodes, return -1 if p<q ,0 if p==q ,1 if p>q.
     1. anything < NUM
     2. x < y
     3. x^3 < x^2 < x < y^3
   note: 3 < 2   ???
   other operators have same order as left operand.
   Historical note... this function WAS used for both sorting and
   testing for equality.  However, for maintenance and efficiency
   reasons I found it better to make a separate "equal" function.
*)
Function cmp(p,q:PNode):Integer;
Var r,i : Integer;
    j,k : Integer;
Begin
 if p^.kind=_NUM then
  begin   { p=nombre }
   if q^.kind=_NUM then
    begin  { q non nombre }
     if p^.value=q^.value then
      cmp:=0
     else if p^.value < q^.value then
      cmp:=1
     else cmp:=-1;
     exit;
    end
   else
    begin
     cmp:=1;
     exit;
    end;
  end
 else if q^.kind=_NUM then
  begin   { p non nombre et q nombre }
   cmp:=-1;
   exit;
  end;
 j:=Integer((p^.kind=_MUL) and (p^.parm[1]^.kind=_NUM));
 k:=Integer((q^.kind=_MUL) and (q^.parm[1]^.kind=_NUM));
 if (j<>0) and (k=0) then
  begin
   r:=cmp(p^.parm[0],q);     { ignorer le coefficient de p }
   if r<>0 then
    cmp:=r
   else cmp:=1;
   exit;
  end;
 if (k<>0) and (j=0) then
  begin
   r:=cmp(p,q^.parm[0]);     { ignorer le coefficient de q }
   if r<>0 then
    cmp:=r
   else cmp:=-1;
   exit;
  end;
 j:=Integer((p^.kind=_EXP) and (p^.parm[1]^.kind=_NUM));
 k:=Integer((q^.kind=_EXP) and (q^.parm[1]^.kind=_NUM));
 if (j<>0) and (k=0) then
  begin
   r:=cmp(p^.parm[0],q);
   if r<>0 then
    begin
     cmp:=r;
     exit;
    end;
  end;
 if (k<>0) and (j=0) then
  begin
   r:=cmp(p,q^.parm[0]);
   if r<>0 then
    begin
     cmp:=r;
     exit;
    end;
  end;
 if p^.kind=_VAR then
  begin   { p=variable }
   if q^.kind=_VAR then
    begin  { q=variable }
     if p^.name=q^.name then
      cmp:=0
     else if p^.name>q^.name then
      cmp:=1
     else cmp:=-1;
     exit;
    end
   else if q^.nump<>0 then
    begin
     r:=cmp(p,q^.parm[0]);
     if r<>0 then
      begin
       cmp:=r; exit;
      end;
    end;
   cmp:=1;
   exit;
  end
 else if q^.kind=_VAR then
  begin
   if p^.nump<>0 then
    begin
     r:=cmp(p^.parm[0],q);
     if r<>0 then
      begin
       cmp:=r; exit;
      end;
    end;
   cmp:=-1;
   exit;
  end;
 if (p^.kind=_FUN) and (q^.kind=_FUN) then
  begin  { fonctions }
   if p^.name=q^.name then
    r:=0
   else if p^.name>q^.name then
    r:=1
   else r:=-1;
   if r<>0 then
    begin
     cmp:=r;
     exit;
    end;
  end;
 k:=q^.kind-p^.kind;
 if k=0 then k:=p^.nump-q^.nump;
 if k=0 then
  begin   { y a-t-il galit }
   for i:=0 to p^.nump-1 do
    begin
     r:=cmp(p^.parm[i],q^.parm[i]);
     if r<>0 then
      begin
       cmp:=r;
       exit;
      end;
    end;
   cmp:=0;             { galit }
   exit;
  end;
 if (p^.nump<>0) and (q^.nump<>0) then
  begin      { oprateurs diffrents }
   if k>0 then
    r:= cmp(p^.parm[0],q)
   else r:= cmp(p,q^.parm[0]);
   if r<>0 then
    begin
     cmp:=r;
     exit;
    end;
  end;
 if k>0 then cmp:=1 else cmp:=-1;
End;

(*
   sort over operator.  oper = ADD or MUL.
   This uses bubble sort.  It assumes that the nodes have been
   adjusted to left association.  For best results, call bisect
   before calling this.
*)
Function SortNode(p:PNode; oper:Integer):Integer;
Type PPNode=^PNode;
Var i,r:Integer;
    s:PNode;
    t:PPNode;
Begin
 r:=0;
 if p^.kind<>oper then
  begin
   for i:=0 to p^.nump-1 do r:=r+SortNode(p^.parm[i],oper);
   SortNode:=r;
   exit;
  end;
 while p^.kind=oper do
  begin
   s:=p;
   t:=@(p^.parm[1]);
   while s^.parm[0]^.kind=oper do
    begin
     s:=s^.parm[0];
     if cmp(s^.parm[1],t^)>0 then
      t:=@(s^.parm[1]);
    end;
   if cmp(s^.parm[0],t^)>0 then
    t:=@(s^.parm[0]);
   if t^<>p^.parm[1] then
    begin             (* too much indirection? *)
     s:=t^;
     t^:=p^.parm[1];
     p^.parm[1]:=s;
     inc(r);
    end;
   r:=r+SortNode(p^.parm[1],oper);
   p:=p^.parm[0];
  end;
 r:=r+SortNode(p,oper);
 SortNode:=r;
End;

END.

{                       Fin du fichier ALGSORT.PAS                          }
