Unit AlgParse;
{ interprtation d'une chaine de caractres }
{ K.B. 1997 }

INTERFACE

Uses AlgConst, AlgUtil, AlgNode;

Const
 ParseErrorFlag:Integer=0;
 errPile       = 1;  { nombre manquant dans pile }
 errNombre     = 2;  { nombre mal crit }
 errMoins      = 3;  { moins unaire }
 errMem        = 4;  { mmoire insuffisante }
 errOpAttendue = 5;  { opration attendue }
 errSyntaxe    = 6;  { erreur de syntaxe }
 errExpAttendue= 7;  { expression attendue }
 errParenthese = 8;  { erreur de parenthse }
 errTropGrand  = 9;  { nombre trop grand }

Type
 TParseRec=record  { enregistrement pour interprtation }
  S:String;
  P:Integer;
  end;

Function ParseStr(S:String): PNode;
{ interprte une chane }

Procedure push(tt:PNode; var pile:PNode);
{ empile tt sur pile }

Procedure pop(var tt,pile:PNode);
{ retire tt de la pile }

Function ParseErrorMsg:String;

IMPLEMENTATION

Function ParseErrorMsg:String;
Begin
 case ParseErrorFlag of
 errPile        : ParseErrorMsg:='Elment manquant dans la pile.';
 errNombre      : ParseErrorMsg:='Nombre mal crit.';
 errMoins       : ParseErrorMsg:='Moins unaire.';
 errMem         : ParseErrorMsg:='Mmoire insuffisante.';
 errOpAttendue  : ParseErrorMsg:='Opration attendue.';
 errExpAttendue : ParseErrorMsg:='Expression attendue.';
 errSyntaxe     : ParseErrorMsg:='Erreur de syntaxe.';
 errParenthese  : ParseErrorMsg:='Erreur de parenthse.';
 errTropGrand   : ParseErrorMsg:='Nombre trop grand.';
 else ParseErrorMsg:='???';
 end;
End;

Procedure Push(tt:PNode; var pile:PNode);
Begin
 tt^.next := pile;
 pile := tt;
End;

Procedure Pop(var tt,pile:PNode);
Begin
 tt:=pile;
 if pile=nil
  then ParseErrorFlag:=errPile
  else pile:=pile^.next;
End;

Procedure LiberePile(pile:PNode);
Var temp,node:PNode;
Begin
 node:=pile;
 while node<>nil do
  begin
   temp:=node^.next;
   freetree(node);
   node:=temp;
  end;
End;

Procedure ReadNumber(var PR:TParseRec; var t:Real);
Var err:Integer;
    c:Char;
    Buf:TStr80;
Begin
 Buf:='';
 repeat
  c:=PR.S[PR.P];
  if c in ['0'..'9','.'] then
   begin
    Buf:=Buf+c;
    inc(PR.P);
   end;
 until not (c in ['0'..'9','.']);
 if PR.P>0 then dec(PR.P);
 val(Buf,t,err);
 if err<>0 then
  begin
   ParseErrorFlag:=errNombre;
   t:=0;
  end;
 if abs(t)>MaxDouble then
  begin
   ParseErrorFlag:=errTropGrand;
   t:=0;
  end;
End;

(*
   renvoie un lment de l'expression  valuer
   si lireop=true, un oprateur est attendu, sinon c'est
   un nombre, une variable ou une fonction
   if nn:=1 then negative signs are treated as operators.
*)
Function GetToken(var PR:TParseRec; var lirop:Boolean; var pildat:PNode):PNode;
Const Buf:TStr80='';
Var i:Integer;
    n,d:TDouble;
    p: pnode;
    c:Char;
    r:Real;
Begin
 c:=PR.S[PR.P];
 if lirop then
  begin
   case c of
     '*' : gettoken:=newoper(_MUL);
     '/' : gettoken:=newoper(_DIV);
     '-' : gettoken:=newoper(_SUB);
     '+' : gettoken:=newoper(_ADD);
     '^' : gettoken:=newoper(_EXP);
     #0  : gettoken:=nil;
     '=' : gettoken:=newoper(_EQU);
     ',',';' : gettoken:=newoper(_CMA);
     ')' : gettoken:=newoper(_EPA);
     else begin
           ParseErrorFlag:=errOpAttendue;
           GetToken:=nil;
          end;
    end;
   {lirop:=false;}
   exit;
  end;
 case c of
  #0 :
   begin ParseErrorFlag:=errExpAttendue; gettoken:=nil; exit; end;
  '(':
   begin gettoken:=newoper(_SPA); exit; end;
  '-','+':
   begin
    if PR.S[PR.P+1] in ['0'..'9'] then
     begin
      inc(PR.P);
      ReadNumber(PR,r);
      if ParseErrorFlag=0 then
       begin
        if c='+' then
         begin
          if r=round(r) then
           GetToken:=newnum(round(r))
          else GetToken:=newreal(r)
         end
        else
         begin
          if r=round(r) then
           GetToken:=newnum(-round(r))
          else GetToken:=newreal(-r)
         end;
       end
      else GetToken:=nil;
     end
    else
     begin
      if c='-' then
       push(newnum(-1),pildat)
      else push(newnum(1),pildat);
      gettoken:=newoper(_MUL);
      lirop:=true;
     end;
    exit;
   end;
  '0'..'9':
   begin
    Readnumber(PR,r);
    if ParseErrorFlag=0 then
     begin
      if r=round(r) then
       Gettoken:=newnum(round(r))
      else GetToken:=newreal(r);
     end
    else GetToken:=nil;
    exit;
   end;
  else begin
        i:=1;
        buf:=c;
        while (i<25) and (c>' ') do
         begin
          inc(PR.P);
          c:=PR.S[PR.P];
          if pos(c,':=()+-*/^,')<>0 then
           begin
            i:=25;
            if c='(' then
             begin
              p:=newnode;
              p^.kind:=_FUN;
              p^.name:=buf;
              p^.nump := 1;
              gettoken:=p;
              exit;
             end;
           end
          else if c>' ' then
           buf:=buf+c;
          inc(i);
         end;
        if PR.P>0 then dec(PR.P);
        if buf='e' then
         begin gettoken:=newvar(ename); exit; end
        else if buf='pi' then
         begin gettoken:=newvar('pi'); exit; end
        else gettoken:=newvar(buf);
       end;
   end;
End;

Function ParseStr(S:String): PNode;
Var
 PR:TParseRec;
 p,t : PNode;
 pildat,pilop : PNode;
 i : Integer;
 lirop : Boolean;    { indique qu'on attend un oprateur }
 continue : Boolean;
Begin
 ParseErrorFlag:=0;
 PR.S:=S+#0;
 PR.P:=1;
 pildat := nil;
 pilop  := nil;
 lirop  := false;
 p:=gettoken(PR,lirop,pildat);
 continue:=true;
 while (p<>nil) and (ParseErrorFlag=0) do
  begin
   if lirop and ((rpr[p^.kind]=8) or (rpr[p^.kind]=1)) then
    begin        (* dbut d'une nouvelle expression *)
     while (ParseErrorFlag=0) and (pilop<>nil) do  (* on rduit les oprations en cours *)
      begin
       pop(t,pilop);
       if (ParseErrorFlag=0) and (t^.kind<=_EXP) then
        begin
         t^.nump := 2;
         pop(t^.parm[1],pildat);
         if ParseErrorFlag=0
          then begin
                pop(t^.parm[0],pildat);
                if ParseErrorFlag=0 then push(t,pildat);
               end;
        end
        else ParseErrorFlag:=errSyntaxe;
      end;
    end;
   if rpr[p^.kind]=1 then
    begin         (* fonction ou parenthse ouvrante *)
     push(p,pilop);
     lirop:=false;
    end
   else if rpr[p^.kind]=8 then
    begin           (* variable ou nombre *)
     push(p,pildat);
     lirop:=true;
    end
   else if (pilop<>nil) and (rpr[p^.kind] <= rpr[pilop^.kind])
                    and (p^.kind<>_EXP) then
    begin   (* rduire opration de moindre priorit *)
     pop(t,pilop);
     continue:=t^.kind<=_EXP;
     if continue then
      begin
       t^.nump := 2;
       pop(t^.parm[1],pildat);
       if ParseErrorFlag=0
        then begin
              pop(t^.parm[0],pildat);
              if ParseErrorFlag=0 then push(t,pildat);
             end;
       continue:=false;  (* don't get more input *)
      end;
    end
   else if continue and (pilop<>nil) and (pilop^.kind=_FUN)
                    and (p^.kind=_CMA) then
    begin    (* virgule : nouveau paramtre *)
     inc(pilop^.nump);
     freenode(p);
     lirop:=false;
    end
   else if continue and (pilop<>nil) and (pilop^.kind=_FUN)
                    and (p^.kind=_EPA) then
    begin   (* parenthse fermante de fonction *)
     pop(t,pilop);
     if t^.nump>MAXP then t^.nump := MAXP;
     for i:=t^.nump-1 downto 0 do pop(t^.parm[i],pildat);
     push(t,pildat);
     freenode(p);   (* free close paren *)
     lirop:=true;
    end
   else if continue and (pilop<>nil) and (pilop^.kind=_SPA)
                    and (p^.kind=_EPA) then
    begin   (* parenthse fermante *)
     pop(t,pilop);
     freenode(t);
     freenode(p);
     lirop:=true;
    end
   else
    begin
     if continue and (p^.kind > _EXP) then
      begin
       ParseErrorFlag:=errSyntaxe;
      end
     else if continue then
      push(p,pilop); (* SHIFT *)
     if continue then lirop:=false;
    end;
   {lecture de l'lment suivant}
   if continue then
    begin
     inc(PR.P);
     p:=gettoken(PR,lirop,pildat);
    end;
   continue:=true;
  end;  { of while }
 continue:=true;
 while (ParseErrorFlag=0) and (pilop<>nil) do
  begin                       (* REDUCE all remaining oper *)
   pop(t,pilop);
   continue:= t^.kind<=_EXP;
   if continue then
    begin
     t^.nump := 2;
     pop(t^.parm[1],pildat);
     if ParseErrorFlag=0
      then begin
            pop(t^.parm[0],pildat);
            if ParseErrorFlag=0 then push(t,pildat);
           end;
    end
   else begin
         if t^.kind=_SPA
          then ParseErrorFlag:=errParenthese
          else ParseErrorFlag:=errSyntaxe;
         freetree(t);
        end;
  end;
 if ParseErrorFlag=0
  then ParseStr:=pildat
  else begin
        ParseStr:=nil;
        LiberePile(pildat);
        LiberePile(pilop);
       end;
End;

END.

{                         Fin du fichier AlgParse.Pas                      }

