Unit UPCalc;
{ unit de calcul avec interception des erreurs }

INTERFACE

Uses SysUtils;

Const
 MaxPow = 100;

 mRadian = 0;
 mDegre  = 1;
 ModeAngle:Integer=mRadian;

Var
 PCalcErrCode : Word;     { code d'erreur }
 PCalcLastErr : Word;

{ gestions des erreurs }
Function  GetPCalcErrorMsg:String;  { renvoie le message d'erreur }
Function  IsPCalcError:Boolean;     { indique la prsence d'une erreur }

{ oprations arithmtiques }
Function PAdd(x1,x2:Real):Real;
Function PSub(x1,x2:Real):Real;
Function PMul(x1,x2:Real):Real;
Function PDiv(x1,x2:Real):Real;
Function POpp(x:Real):Real;

{ fonctions d'une variable relle prdfinies }
Function PArcTan(x:Real):Real;    { arc tangente }
Function PCos(x:Real):Real;       { cosinus }
Function PSin(x:Real):Real;       { sinus }
Function PTan(x:Real):Real;       { tangente }
Function PLn(x:Real):Real;        { logarithme }
Function PExp(x:Real):Real;       { exponentielle }
Function PSqr(x:Real):Real;       { carr }
Function PSqrt(x:Real):Real;      { racine carre }

{ autres fonctions }
Function PSh(x:Real):Real;        { sinus hyperbolique }
Function PCh(x:Real):Real;        { cosinus hyperbolique }
Function PTh(x:Real):Real;        { tangente hyperbolique }
Function PArcSin(x:Real):Real;    { arc sinus }
Function PArcCos(x:Real):Real;    { arc cosinus }
Function PArgSh(x:Real):Real;     { argument sinus hyperbolique }
Function PArgCh(x:Real):Real;     { argument cosinus hyperbolique }
Function PArgTh(x:Real):Real;     { argument tangente hyperbolique }
Function PLog(x:Real):Real;       { logarithme dcimal }
Function PPuiss(x1,x2:Real):Real; { x1 puissance x2 }
Function PRound(x:Real):Real;     { arrondi entier de x }
Function PFrac(x:Real):Real;      { partie fractionnaire de x }
Function PINT(x:Real):Real;       { partie entire de x }
Function PABS(x:Real):Real;       { valeur absolue de x }
Function PRANDOM(x:Real):Real;    { nombre tir au hasard infrieur  x }

Function PHypot(x,y:Real):Real;    { hypotnuse }
Function PArcTan2(x,y:Real):Real;  { angle }

Function PMod(x,y:Real):Real;     { reste de la division de x par y }
Function PSign(x:Real):Real;      { signe de x }
Function PMin(x,y:Real):Real;     { minimum }
Function PMax(x,y:Real):Real;     { maximum }

Function Rad2Deg(x:Real):Real;
Function Deg2Rad(x:Real):Real;

Procedure ModeDegre;

IMPLEMENTATION

Const
 cInvalidOp=1;
 cZeroDivide=2;
 cOverflow=3;
 cUnderflow=4;

Procedure ModeDegre;
Begin
 ModeAngle:=mDegre;
End;

Function GetPCalcErrorMsg:String;
Begin
 case PCalcLastErr of
    cUnderflow  : GetPCalcErrorMsg:='Dpassement de capacit.';
    cZeroDivide : GetPCalcErrorMsg:='Division par zro.';
    cOverflow   : GetPCalcErrorMsg:='Dpassement de capacit .';
    cInvalidOp  : GetPCalcErrorMsg:='Opration non valide.';
    else GetPCalcErrorMsg:='R.A.S';
 end;
End;

Function IsPCalcError:Boolean;
Begin
 PCalcLastErr:=PCalcErrCode;
 PCalcErrCode:=0;
 IsPCalcError:=PCalcLastErr<>0;
End;

{ conversions angulaires }

Function Rad2Deg(x:Real):Real;
Begin
 Rad2Deg:=PDiv(PMul(x,180),pi);
End;

Function Deg2Rad(x:Real):Real;
Begin
 Deg2Rad:=PDiv(PMul(x,pi),180);
End;

{ oprations arithmtiques }

Function PAdd(x1,x2:Real):Real;
Begin
 try
  Result:=x1+x2;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
 end;
End;

Function PSub(x1,x2:Real):Real;
Begin
 try
  Result:=x1-x2;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
 end;
End;

Function PMul(x1,x2:Real):Real;
Begin
 try
  Result:=x1*x2;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
 end;
End;

Function PDiv(x1,x2:Real):Real;
Begin
 try
  if x2<>0 then Result:=x1/x2
  else
   begin
    PCalcErrCode:=cZeroDivide;
    Result:=0;
   end;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
 end;
End;

Function POpp(x:Real):Real;
Begin
 Result:=-x;
End;

{ fonctions d'une variable relle prdfinies }

Function PArcTan(x:Real):Real;    { arc tangente }
Var t:Real;
Begin
 try
  t:=ArcTan(x);
  case ModeAngle of
   mRadian : Result:=t;
   else Result:=Rad2Deg(t);
  end;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PCos(x:Real):Real;       { cosinus }
Var t:Real;
Begin
 try
  case ModeAngle of
   mRadian : t:=x;
   else t:=Deg2Rad(x);
   end;
  Result:=cos(t);
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PSin(x:Real):Real;       { sinus }
Var t:Real;
Begin
 try
  case ModeAngle of
   mRadian : t:=x;
   else t:=Deg2Rad(x);
   end;
  Result:=Sin(t);
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PLn(x:Real):Real;        { logarithme }
Begin
 try
  if x>0 then Result:=Ln(x)
  else
   begin
    PCalcErrCode:=cInvalidOp;
    Result:=0;
   end;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
 end;
End;

Function PExp(x:Real):Real;       { exponentielle }
Begin
 try
  Result:=Exp(x);
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PSqr(x:Real):Real;       { carr }
Begin
 try
  Result:=Sqr(x);
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PSqrt(x:Real):Real;      { racine carre }
Begin
 try
  if x>=0 then Result:=Sqrt(x)
  else
   begin
    PCalcErrCode:=cInvalidOp;
    Result:=0;
   end;
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

{ autres fonctions }

Function PTan(x:Real):Real;       { tangente }
Begin
 PTan:=PDiv(PSin(x),PCos(x));
End;

Function PCh(x:Real):Real;        { cosinus hyperbolique }
Begin
 PCh:=PDiv(PAdd(PExp(x),PExp(-x)),2);
End;

Function PSh(x:Real):Real;        { sinus hyperbolique }
Begin
 PSh:=PDiv(PSub(PExp(x),PExp(-x)),2);
End;

Function PTh(x:Real):Real;        { tangente hyperbolique }
Var r1,r2:Real;
Begin
 r1:=PSub(PExp(x),PExp(-x));
 r2:=PAdd(PExp(x),PExp(-x));
 PTh:=PDiv(r1,r2);
End;

Function PArcSin(x:Real):Real;    { arc sinus }
Begin
 if x=1 then
  begin
   case ModeAngle of
    mRadian : PArcSin:=pi/2;
    else PArcSin:=90;
   end;
   exit;
  end;
 if x=-1 then
  begin
   case ModeAngle of
    mRadian : PArcSin:=-pi/2;
    else PArcSin:=-90;
   end;
   exit;
  end;
 PArcSin:=PArcTan(PDiv(x,PSqrt(PSub(1,PSqr(x)))));
End;

Function PArcCos(x:Real):Real;    { arc cosinus }
Var t:Real;
Begin
 if x=0 then
  begin
   case ModeAngle of
    mRadian : PArcCos:=pi/2;
    else PArcCos:=90;
   end;
   exit;
  end;
 t:=PArcTan(PDiv(PSqrt(PSub(1,PSqr(x))),x));
 if x<0 then
  begin
   case ModeAngle of
    mRadian : t:=PAdd(pi,t);
    else t:=PAdd(180,t);
    end;
   end;
  PArcCos:=t;
End;

Function PArgSh(x:Real):Real;     { argument sinus hyperbolique }
Var r:Real;
Begin
 r:=PSqrt(PAdd(PSqr(x),1));
 if x<0
    then PArgSh:=PLn(PSub(r,x))
    else PArgSh:=PLn(PAdd(x,r));
End;

Function PArgCh(x:Real):Real;     { argument cosinus hyperbolique }
Begin
 PArgCh:=PLn(PAdd(x,PSqrt(PSub(PSqr(x),1))));
End;

Function PArgTh(x:Real):Real;     { argument tangente hyperbolique }
Begin
 PArgTh:=PMul(0.5,PLn(PDiv(PAdd(1,x),PSub(1,x))));
End;

Function PLog(x:Real):Real;       { logarithme dcimal }
Begin
 PLog:=PDiv(PLn(x),PLn(10));
End;

Function PPuiss(x1,x2:Real):Real;     { x1 puissance x2 }
Var k,n:Integer;
    r:Real;
Begin
 if (PFrac(x2)=0) and (abs(x2)<MaxPow) then
  begin
   n:=round(x2);
   k:=0;
   r:=1;
   while (k<abs(n)) and (PCalcErrCode=0) do
    begin
     r:=PMul(r,x1);
     inc(k);
    end;
   if PCalcErrCode=0 then
    begin
     if n>0 then PPuiss:=r
     else PPuiss:=PDiv(1,r);
    end
   else PPuiss:=0;
  end
 else PPuiss:=PExp(PMul(x2,PLn(x1)));
End;

Function PRound(x:Real):Real;
Begin
 try
  Result:=Round(x);
 except
  on EOverflow do
   begin Result:=0; PCalcErrCode:=cOverflow; end;
  on EUnderflow do
   begin Result:=0; PCalcErrCode:=cUnderflow; end;
  on EZeroDivide do
   begin Result:=0; PCalcErrCode:=cZeroDivide; end;
  on EInvalidOp do
   begin Result:=0; PCalcErrCode:=cInvalidOp; end;
 end;
End;

Function PFrac(x:Real):Real;
Begin
 PFrac:=Frac(x);
End;

Function PINT(x:Real):Real;
Begin
 if x>=0 then PINT:=Int(x)
 else PINT:=Int(x)-1;
End;

Function PABS(x:Real):Real;
Begin
 PABS:=Abs(x);
End;

Function PRANDOM(x:Real):Real;
Begin
 PRANDOM:=PMUL(x,random);
End;

Function PHypot(x,y:Real):Real;  { hypotnuse }
Begin
 PHypot:=PSqrt(PAdd(PSqr(x),PSqr(y)));
End;

Function PArcTan2(x,y:Real):Real;  { angle }
Var t:Real;
Begin
 if x=0 then t:=pi/2
 else t:=PArcTan(PDiv(y,x));
 case ModeAngle of
  mRadian : PArcTan2:=t;
  else PArcTan2:=Rad2Deg(t);
  end;
End;

Function PMod(x,y:Real):Real;
Begin
 if y=0 then PMod:=x
 else PMod:=PSub(x,PMul(PRound(PDiv(x,y)),y));
End;

Function PSign(x:Real):Real;
Begin
 if x=0 then PSign:=0
 else if x>0 then PSign:=1 else PSign:=-1;
End;

Function PMin(x,y:Real):Real;
Begin
 if x<y then PMin:=x else PMin:=y;
End;

Function PMax(x,y:Real):Real;
Begin
 if x>y then PMax:=x else PMax:=y;
End;

END.

{                          Fin du fichier UPCALC.PAS                        }
