unit Umain;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, Clipbrd, ShellAPI,
  AlgConst, AlgUtil, AlgNode, AlgParse, AlgExp, AlgCalc, AlgAppr, AlgShow,
  Buttons, About;


type
  TForm1 = class(TForm)
    Panel1: TPanel;
    InsButton: TSpeedButton;
    CopyButton: TSpeedButton;
    CalcButton: TSpeedButton;
    EnvButton: TSpeedButton;
    HelpButton: TSpeedButton;
    AboutButton: TSpeedButton;
    ExitButton: TSpeedButton;
    Label1: TLabel;
    Edit1: TEdit;
    Image1: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure InsButtonClick(Sender: TObject);
    procedure CalcButtonClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Edit1KeyPress(Sender: TObject; var Key: Char);
    procedure AboutButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure EnvButtonClick(Sender: TObject);
    procedure ExitButtonClick(Sender: TObject);
    procedure HelpButtonClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
  private
    { Private-dclarations }
  public
    { Public-dclarations }
    Node: PNode;          { la formule affiche }
    procedure ShowNode;   { affichage de la formule }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
 node:=nil; { pas de formule au dpart }
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 { librer la mmoire }
 if node<>nil then freetree(node);
end;

procedure TForm1.InsButtonClick(Sender: TObject);
begin
 { activer l'dition d'une nouvelle formule }
 Edit1.Selstart:=0;
 Edit1.SelLength:=length(Edit1.Text);
 Edit1.SetFocus;
end;

procedure TForm1.CalcButtonClick(Sender: TObject);
{ bouton Calcul algbrique }
var p1,p2: PNode;
    i: Integer;
begin
 {y a-t-il quelque chose  calculer ? }
 if node=nil then exit;
 {on ne calcule pas les galits }
 if (node^.kind=_EQU) or (node^.kind=_ENV) then exit;
 { pour les nombres entiers on dcompose en facteurs premiers }
 if node^.kind=_NUM then
  begin
   p1:=deepcopy(node);
   p2:=deepcopy(node);
   PrimeFact(p2);
  end
 else
 { sinon on effectue le calcul algbrique }
  begin
   p2:=AlgEval(node);
   p1:=deepcopy(node);
  end;
 { on efface la formule prcdente et on met la nouvelle qui contient
   le rsultat obtenu }
 if node^.nump>0 then
  for i:=0 to node^.nump-1 do freetree(node^.parm[i]);
 node^.kind:=_EQU;
 node^.nump:=2;
 node^.parm[0]:=p1;
 node^.parm[1]:=p2;
 ShowNode;
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
{ touche F1 : aide }
if key=vk_F1 then
 begin
  HelpButtonClick(Sender);
  key:=0;
 end;
if Sender=Edit1 then exit;
{ touche Inser : on active la ligne d'dition }
if key=vk_Insert then
 begin
  InsButtonClick(Sender);
  key:=0;
 end;
{ touche = : calcul algbrique ou calcul approch si avec touche Ctrl }
if key=187 {=} then
 begin
  if ssCtrl in Shift then
   EnvButtonClick(Sender)       { calcul approch }
  else CalcButtonClick(Sender); { calcul algbrique }
  key:=0;
 end;
end;

procedure TForm1.ShowNode;
{ affichage de la formule dans un bitmap noir et blanc }
var bitmap:TBitmap;
begin
if node=nil then exit;
bitmap:=TBitmap.Create;
bitmap.monochrome:=true;
with bitmap.canvas do
 begin
  font.name:='Courier New';
  font.size:=10;
 end;
{ calculer les dimensions et les positions }
resizenode(bitmap.canvas,node,0);
{ donner les dimensions correctes au bitmap : marge de 5 pixels }
bitmap.width:=node^.sx+10;
bitmap.height:=node^.dyh+node^.dyb+10;
{ afficher }
affnode(bitmap.canvas,node,5,5);
image1.picture.assign(bitmap);
end;

procedure TForm1.Edit1KeyPress(Sender: TObject; var Key: Char);
{ appui sur Entre dans la ligne d'dition : on interprte la ligne
  de saisie et on affiche la formule }
var S:String;
    p:PNode;
begin
if Key=#13 then
 begin
  Key:=#0;
  S:=Edit1.Text;
  p:=ParseStr(S);
  if ParseErrorFlag<>0 then
   begin { erreur dans l'interprtation }
    S:=ParseErrorMsg+#0;
    Application.MessageBox(@S[1],'Erreur',idOk);
   end
  else
   begin { pas d'erreur }
    if node<>nil then freetree(node);
    node:=p;
    Shownode;
   end;
 end;
end;

procedure TForm1.AboutButtonClick(Sender: TObject);
{ bouton A propos }
begin
 AboutBox.ShowModal;
end;

procedure TForm1.CopyButtonClick(Sender: TObject);
{ copie dans le Presse-Papiers }
begin
if node<>nil then
 ClipBoard.Assign(Image1.Picture);
end;

procedure TForm1.EnvButtonClick(Sender: TObject);
{ calcul approch de la formule }
var p:PNode;
    rep:Real;
begin
 if node=nil then exit;
 if EvalNode(node,rep) then
  begin
   p:=deepcopy(node);
   MoveNode(node,cons(p,_ENV,NewReal(rep)));
   ShowNode;
  end;
end;

procedure TForm1.ExitButtonClick(Sender: TObject);
{ bouton Quitter }
begin
 close;
end;

procedure TForm1.HelpButtonClick(Sender: TObject);
{ bouton Aide }
begin
 Application.HelpJump('Main_Index');
end;

procedure TForm1.FormResize(Sender: TObject);
{ changement de dimensions de la fentre : on ajuste la largeur et la
  hauteur de la zone d'affichage de la formule }
begin
 Image1.Height:=Height-Panel1.Height;
 Image1.Width:=Width;
 ShowNode;
end;


End.
